import pathlib
import shutil

from dict_tools.data import NamespaceDict


def make_module(path: pathlib.Path):
    """
    Recursively add an empty __init__.py file to this dir and subdirs recursively.
    Otherwise pytest will have issues with duplicate names of directories and files.
    """
    if path.is_dir():
        init = path / "__init__.py"
        init.touch()
        for p in path.iterdir():
            make_module(p)


if __name__ == "__main__":
    ctx = NamespaceDict({{cookiecutter}})

    root_directory = pathlib.Path.cwd()
    test_dir = root_directory / "tests"

    if ctx.vertical:
        # Vertically app-merged projects won't have an entrypoint for this test
        for test_type in ("integration", "unit"):
            shutil.rmtree(test_dir / test_type / ctx.clean_name)

    # Add __init__.py files under every directory in the "tests" dir
    make_module(test_dir)



import pathlib

from dict_tools.data import NamespaceDict

if __name__ == "__main__":
    ctx = NamespaceDict({{cookiecutter}})
    root_directory = pathlib.Path.cwd()

    for dyne in ctx.dyne_list:
        contract = root_directory / ctx.clean_name / dyne / "contracts"
        try:
            contract.mkdir(parents=True)
        except FileExistsError:
            ...



def context(hub, ctx, directory: str):
    ctx.version_number = "0.1.0"

    ctx.welcome = f"Welcome to {ctx.project_name}'s Documentation!"
    ctx.welcome = f"{ctx.welcome}\n{'=' * len(ctx.welcome)}\n"

    return ctx



#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config
# -- Path setup --------------------------------------------------------------
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
import datetime
from pathlib import Path

# -- Project information -----------------------------------------------------
this_year = datetime.datetime.today().year
if this_year == {{cookiecutter.this_year}}:
    copyright_year = {{cookiecutter.this_year}}
else:
    copyright_year = f"{{cookiecutter.this_year}} - {this_year}"
project = "{{cookiecutter.project_name}}"
copyright = f"{copyright_year}, {{cookiecutter.author}}"
author = "{{cookiecutter.author}}"

# Strip version info from ../../{{cookiecutter.clean_name}}/version.py
with open(Path(Path(__file__).parent.parent, '{{cookiecutter.clean_name}}', 'version.py')) as version_file:
    content = version_file.readlines()
    for file_line in content:
        if 'version =' in file_line:
            version = file_line.split(' ')[2].replace('"','')
            break

# Variables to pass into the docs from sitevars.rst for rst substitution
with open("sitevars.rst") as site_vars_file:
    site_vars = site_vars_file.read().splitlines()

rst_prolog = """
{}
""".format(
    "\n".join(site_vars[:])
)

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
needs_sphinx = "3.5.3"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx_copybutton",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    # "sphinxcontrib.spelling",
]
# Render TODO directives, set to FALSE before publishing
# This is incredibly helpful, when set to True, to know what is yet to be
# completed in documentation.
todo_include_todos = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    ".vscode",
    ".venv",
    ".git",
    ".gitlab-ci",
    ".gitignore",
    "sitevars.rst",
]

autosummary_generate = True

# ----- Napolean Config ------------------------------------------------------
# For using Google-style docstrings in Python code as a standard, which is
# highly recommended. This improves tooling by expecting a standard way of
# using docstrings in your project.
# https://www.sphinx-doc.org/en/master/usage/extensions/napoleon.html
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True

# ----- Intersphinx Config ---------------------------------------------------
# This extension can generate automatic links to the documentation of objects
# in other projects, such as the official Python or POP docs.
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "pytest": ("https://pytest.readthedocs.io/en/stable", None),
    "pop": ("https://pop.readthedocs.io/en/latest/", None),
}

# ----- Autodoc Config -------------------------------------------------------
# This extension can import the modules you are documenting, and pull in
# documentation from docstrings in a semi-automatic way. This is powerful!
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html
autodoc_default_options = {"member-order": "bysource"}

# ----- Autosummary Config ---------------------------------------------------
# This extension generates function/method/attribute summary lists, similar to
# those output e.g. by Epydoc and other API doc generation tools. This is
# especially useful when your docstrings are long and detailed, and putting
# each one of them on a separate page makes them easier to read.
# https://www.sphinx-doc.org/en/master/usage/extensions/autosummary.html
autosummary_generate = True

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"
html_title = f"{project} Documentation"
html_show_sourcelink = True  # False on private repos; True on public repos

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# For example, official Salt Project docs use images from the salt-branding-guide
# https://gitlab.com/saltstack/open/salt-branding-guide/
#
# Example for >=4.0.0 of Sphinx (support for favicon via URL)
# html_logo = "https://gitlab.com/saltstack/open/salt-branding-guide/-/raw/master/logos/SaltProject_altlogo_teal.png?inline=true"
# Example for <4.0.0 of Sphinx, if added into _static/img/ and html_static_path is valid
# html_logo = "_static/img/SaltProject_altlogo_teal.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large. Favicons can be up to at least 228x228. PNG
# format is supported as well, not just .ico'
# For example, official Salt Project docs use images from the salt-branding-guide
# https://gitlab.com/saltstack/open/salt-branding-guide/
#
# Example for >=4.0.0 of Sphinx (support for favicon via URL)
# html_favicon = "https://gitlab.com/saltstack/open/salt-branding-guide/-/raw/master/logos/SaltProject_Logomark_teal.png?inline=true"
# Example for <4.0.0 of Sphinx, if added into _static/img/ and html_static_path is valid
# html_favicon = "_static/img/SaltProject_Logomark_teal.png"

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "{{cookiecutter.clean_name}}doc"


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "{{cookiecutter.clean_name}}.tex",
        "{{cookiecutter.project_name}} Documentation",
        "{{cookiecutter.author}}",
        "manual",
    ),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        master_doc,
        "{{cookiecutter.clean_name}}",
        "{{cookiecutter.project_name}} Documentation",
        [author],
        1,
    )
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "{{cookiecutter.clean_name}}",
        "{{cookiecutter.project_name}} Documentation",
        author,
        "{{cookiecutter.clean_name}}",
        "One line description of project.",
        "Miscellaneous",
    ),
]


# -- Extension configuration -------------------------------------------------



def __init__(hub):
    # Remember not to start your app in the __init__ function
    # This function should just be used to set up the plugin subsystem
    # The run.py is where your app should usually start
    for dyne in {{cookiecutter.short_dyne_list}}:
        hub.pop.sub.add(dyne_name=dyne)


def cli(hub):
    hub.pop.config.load({{cookiecutter.dyne_list}}, cli="{{cookiecutter.clean_name}}")
    # Your app's options can now be found under hub.OPT.{{cookiecutter.clean_name}}
    kwargs = dict(hub.OPT.{{cookiecutter.clean_name}})

    # Initialize the asyncio event loop
    hub.pop.loop.create()

    # Start the async code
    coroutine = hub.{{cookiecutter.clean_name}}.init.run(**kwargs)
    hub.pop.Loop.run_until_complete(coroutine)

async def run(hub, **kwargs):
    """
    This is the entrypoint for the async code in your project
    """
    print("{{cookiecutter.project_name}} works!")



import os

CLI_CONFIG = {
    "config": {"options": ["-c"], "subcommands": ["_global_"]},
    "author": {},
    "author_email": {},
    "project_name": {
        "subcommands": ["_global_"],
        "options": ["-n", "--name"],
    },
    "dyne": {"subcommands": ["_global_"], "options": ["-d"]},
    "overwrite_existing": {
        "options": ["--overwrite", "-o"],
        "subcommands": ["_global_"],
        "action": "store_true",
    },
    "directory": {"options": ["-D"], "subcommands": ["_global_"]},
    "ignore": {"options": ["-I"], "subcommands": ["_global_"]},
    "vertical": {
        "subcommands": ["_global_"],
        "action": "store_true",
        "options": ["-tv"],
    },
}
CONFIG = {
    "config": {
        "default": "",
        "help": "Load extra options from a configuration file",
    },
    "author": {
        "default": None,
        "help": "The name of the author, defaults to git global config value",
    },
    "author_email": {
        "default": None,
        "help": "The author email, defaults to git global config value",
    },
    "project_name": {
        "help": "The name of the project that is being created",
        "default": None,
    },
    "vertical": {
        "default": False,
        "help": "Build a vertically app-merged project, it's entrypoint is in another project",
    },
    "overwrite_existing": {
        "default": False,
        "help": "Overwrite files if they already exist",
    },
    "dyne": {
        "default": [],
        "nargs": "*",
        "help": "A space delimited list of additional dynamic names for vertical app-merging",
    },
    "directory": {
        "default": os.getcwd(),
        "help": "The directory to create the project in",
    },
    "ignore": {
        "default": [],
        "nargs": "*",
        "help": "A space delimited list of file paths to ignore, they won't be changed by pop-create under any circumstances",
    },
}

SUBCOMMANDS = {
    "seed": {"help": "Seed a traditional pop project"},
    "tests": {"help": "Create the tests for a traditional pop-project"},
    "docs": {"help": "Create the Sphinx tooling for this project"},
    "cicd": {"help": "Create the cicd tooling for this project"},
}
DYNE = {
    "pop_create": ["pop_create"],
    "tool": ["tool"],
}



"""
noxfile
~~~~~~~
Nox configuration script
"""
# pylint: disable=resource-leakage,3rd-party-module-not-gated
from pathlib import Path
import datetime
import os
import pathlib
import shutil
import sys

# fmt: off
if __name__ == "__main__":
    sys.stderr.write(
        "Do not execute this file directly. Use nox instead, it will know how to handle this file\n"
    )
    sys.stderr.flush()
    exit(1)
# fmt: on

import nox  # isort:skip
from nox.command import CommandFailed  # isort:skip

# Nox options
#  Reuse existing virtualenvs
nox.options.reuse_existing_virtualenvs = True
#  Don't fail on missing interpreters
nox.options.error_on_missing_interpreters = False

# Python versions to test against
PYTHON_VERSIONS = ("3", "3.6", "3.7", "3.8", "3.9")
# Be verbose when runing under a CI context
CI_RUN = os.environ.get("CI")
PIP_INSTALL_SILENT = CI_RUN is False
SKIP_REQUIREMENTS_INSTALL = "SKIP_REQUIREMENTS_INSTALL" in os.environ
EXTRA_REQUIREMENTS_INSTALL = os.environ.get("EXTRA_REQUIREMENTS_INSTALL")

COVERAGE_VERSION_REQUIREMENT = "coverage==5.5"

# Prevent Python from writing bytecode
os.environ["PYTHONDONTWRITEBYTECODE"] = "1"

# Global Path Definitions
REPO_ROOT = pathlib.Path(__file__).resolve().parent
# Change current directory to REPO_ROOT
os.chdir(REPO_ROOT)

ARTIFACTS_DIR = REPO_ROOT / "artifacts"
# Make sure the artifacts directory exists
ARTIFACTS_DIR.mkdir(parents=True, exist_ok=True)

RUNTESTS_LOGFILE = ARTIFACTS_DIR / "runtests-{}.log".format(
    datetime.datetime.now().strftime("%Y%m%d%H%M%S.%f")
)
COVERAGE_REPORT_DB = REPO_ROOT / ".coverage"
COVERAGE_REPORT_PROJECT = ARTIFACTS_DIR.relative_to(REPO_ROOT) / "coverage-project.xml"
COVERAGE_REPORT_TESTS = ARTIFACTS_DIR.relative_to(REPO_ROOT) / "coverage-tests.xml"
JUNIT_REPORT = ARTIFACTS_DIR.relative_to(REPO_ROOT) / "junit-report.xml"


def _get_session_python_version_info(session):
    try:
        version_info = session._runner._real_python_version_info
    except AttributeError:
        session_py_version = session.run_always(
            "python",
            "-c" 'import sys; sys.stdout.write("{}.{}.{}".format(*sys.version_info))',
            silent=True,
            log=False,
        )
        version_info = tuple(
            int(part) for part in session_py_version.split(".") if part.isdigit()
        )
        session._runner._real_python_version_info = version_info
    return version_info


def _get_pydir(session):
    version_info = _get_session_python_version_info(session)
    if version_info < (3, 6):
        session.error("Only Python >= 3.6 is supported")
    return "py{}.{}".format(*version_info)


@nox.session(python=PYTHON_VERSIONS)
def tests(session):
    if SKIP_REQUIREMENTS_INSTALL is False:
        # Always have the wheel package installed
        session.install("wheel", silent=PIP_INSTALL_SILENT)
        session.install(COVERAGE_VERSION_REQUIREMENT, silent=PIP_INSTALL_SILENT)
        session.install("-e", ".", silent=PIP_INSTALL_SILENT)

        # Install requirements
        requirements_file = (
            REPO_ROOT / "requirements" / _get_pydir(session) / "tests.txt"
        )
        install_command = [
            "--progress-bar=off",
            "-r",
            str(requirements_file.relative_to(REPO_ROOT)),
        ]
        session.install(*install_command, silent=PIP_INSTALL_SILENT)

        if EXTRA_REQUIREMENTS_INSTALL:
            session.log(
                "Installing the following extra requirements because the "
                "EXTRA_REQUIREMENTS_INSTALL environment variable was set: "
                "EXTRA_REQUIREMENTS_INSTALL='%s'",
                EXTRA_REQUIREMENTS_INSTALL,
            )
            install_command = ["--progress-bar=off"]
            install_command += [
                req.strip() for req in EXTRA_REQUIREMENTS_INSTALL.split()
            ]
            session.install(*install_command, silent=PIP_INSTALL_SILENT)

    session.run("coverage", "erase")
    args = [
        "--rootdir",
        str(REPO_ROOT),
        f"--log-file={RUNTESTS_LOGFILE}",
        "--log-file-level=debug",
        "--show-capture=no",
        f"--junitxml={JUNIT_REPORT}",
        "--showlocals",
        "-ra",
        "-s",
    ]
    if session._runner.global_config.forcecolor:
        args.append("--color=yes")
    if not session.posargs:
        args.append("tests/")
    else:
        for arg in session.posargs:
            if arg.startswith("--color") and args[0].startswith("--color"):
                args.pop(0)
            args.append(arg)
        for arg in session.posargs:
            if arg.startswith("-"):
                continue
            if arg.startswith(f"tests{os.sep}"):
                break
            try:
                pathlib.Path(arg).resolve().relative_to(REPO_ROOT / "tests")
                break
            except ValueError:
                continue
        else:
            args.append("tests/")
    try:
        session.run("coverage", "run", "-m", "pytest", *args)
    finally:
        # Always combine and generate the XML coverage report
        try:
            session.run("coverage", "combine")
        except CommandFailed:
            # Sometimes some of the coverage files are corrupt which would
            # trigger a CommandFailed exception
            pass
        # Generate report for salt code coverage
        session.run(
            "coverage",
            "xml",
            "-o",
            str(COVERAGE_REPORT_PROJECT),
            "--omit=tests/*",
            "--include={{cookiecutter.clean_name}}/*",
        )
        # Generate report for tests code coverage
        session.run(
            "coverage",
            "xml",
            "-o",
            str(COVERAGE_REPORT_TESTS),
            "--omit={{cookiecutter.clean_name}}/*",
            "--include=tests/*",
        )
        # Move the coverage DB to artifacts/coverage in order for it to be archived by CI
        if COVERAGE_REPORT_DB.exists():
            shutil.move(
                str(COVERAGE_REPORT_DB), str(ARTIFACTS_DIR / COVERAGE_REPORT_DB.name)
            )


@nox.session(name="docs-html", python="3")
@nox.parametrize("clean", [False, True])
def docs_html(session, clean):
    """
    Build Sphinx HTML Documentation
    """
    pydir = _get_pydir(session)

    # Latest pip, setuptools, and wheel
    install_command = ["--progress-bar=off", "-U", "pip", "setuptools", "wheel"]
    session.install(*install_command, silent=True)

    # Install requirements
    requirements_file = Path("requirements", "docs.txt")
    install_command = ["--progress-bar=off", "-r", str(requirements_file)]
    session.install(*install_command, silent=True)

    build_dir = Path("docs", "_build", "html")
    sphinxopts = "-Wn"
    if clean:
        sphinxopts += "E"
    args = [sphinxopts, "--keep-going", "docs", str(build_dir)]
    session.run("sphinx-build", *args, external=True)


@nox.session(python="3")
def docs(session) -> None:
    """
    Build and serve the Sphinx HTML documentation, with live reloading on file changes, via sphinx-autobuild.

    Note: Only use this in INTERACTIVE DEVELOPMENT MODE. This SHOULD NOT be called
        in CI/CD pipelines, as it will hang.
    """
    pydir = _get_pydir(session)

    # Latest pip, setuptools, and wheel
    install_command = ["--progress-bar=off", "-U", "pip", "setuptools", "wheel"]
    session.install(*install_command, silent=True)

    # Install requirements
    requirements_file = Path("requirements", "docs.txt")
    install_command = ["--progress-bar=off", "-r", str(requirements_file)]
    session.install(*install_command, silent=True)

    # Install autobuild req
    install_command = ["--progress-bar=off", "-U", "sphinx-autobuild"]
    session.install(*install_command, silent=True)

    # Launching LIVE reloading Sphinx session
    build_dir = Path("docs", "_build", "html")
    args = ["--watch", ".", "--open-browser", "docs", str(build_dir)]
    if build_dir.exists():
        shutil.rmtree(build_dir)

    session.run("sphinx-autobuild", *args)



import pathlib
import shutil

from dict_tools.data import NamespaceDict

if __name__ == "__main__":
    ctx = NamespaceDict({{cookiecutter}})
    root_directory = pathlib.Path.cwd()

    if ctx.vertical:
        script = root_directory / ctx.clean_name / "scripts.py"
        script.unlink()

        run = root_directory / "run.py"
        run.unlink()

        non_vertical_dyne = root_directory / ctx.clean_name / ctx.clean_name
        shutil.rmtree(non_vertical_dyne)



import os
import pathlib
import shutil
from typing import Callable


def __init__(hub):
    hub.tool.path.IGNORE = []


def touch(hub, path: pathlib.Path):
    """
    Make sure that a file exists, it's ok if the file already exists
    """
    hub.tool.path.mkdir(path.parent)
    if not path.exists():
        path.touch()


def mkdir(hub, path: pathlib.Path):
    """
    make sure that a directory and it's parents exist
    """
    for p in path.parents:
        # Make sure parent's exist before continuing
        hub.tool.path.mkdir(p)
    if not path.exists():
        path.mkdir()


def copy(hub, src, dst, *, follow_symlinks: bool = True):
    """
    Copy function that skips existing files
    """
    if os.path.isdir(dst):
        dst = os.path.join(dst, os.path.basename(src))
    if str(dst) in hub.tool.path.IGNORE:
        hub.log.debug(f"Ignoring target: {dst}")
        return dst
    elif os.path.isdir(dst) or os.path.isdir(src):
        return dst
    elif os.path.exists(dst):
        hub.log.debug(f"Skipping target that already exists: {dst}")
        return dst
    hub.log.debug(f"Writing new file: {dst}")
    shutil.copyfile(src, dst, follow_symlinks=follow_symlinks)
    shutil.copystat(src, dst, follow_symlinks=follow_symlinks)
    return dst


def copy2(hub, src, dst, *, follow_symlinks: bool = True):
    if str(dst) in hub.tool.path.IGNORE:
        hub.log.debug(f"Ignoring overwrite target: {dst}")
        return dst
    return shutil.copy2(src, dst, follow_symlinks=follow_symlinks)


def copytree(hub, src: pathlib.Path, dst: pathlib.Path, copy_function: Callable):
    if src.is_dir():
        if not dst.exists():
            dst.mkdir()
        for src_subdir in src.iterdir():
            hub.tool.path.copytree(src_subdir, dst / src_subdir.name, copy_function)
    else:
        copy_function(src, dst)


def rmtree(hub, src: pathlib.Path):
    """
    Remove a directory and all files underneath it
    """
    shutil.rmtree(src, ignore_errors=True)
    if src.exists():
        src.rmdir()


def delete(hub, src: pathlib.Path):
    if src.exists():
        src.unlink()



def test_cli(mock_hub, hub):
    mock_hub.{{cookiecutter.clean_name}}.init.cli = hub.{{cookiecutter.clean_name}}.init.cli
    mock_hub.{{cookiecutter.clean_name}}.init.cli()
    mock_hub.pop.config.load.assert_called_once_with(
        ["{{cookiecutter.clean_name}}"], "{{cookiecutter.clean_name}}"
    )



def context(hub, ctx, directory: str):
    if ctx.vertical:
        entrypoint = ""
    else:
        entrypoint = f"{ctx.project_name} = {ctx.clean_name}.scripts:start"

    return dict(
        dyne_dict={f"{d}": [f"{d}"] for d in ctx.dyne_list},
        entrypoint=entrypoint,
        **ctx,
    )



import datetime
import pathlib
import subprocess
import tempfile
from typing import List

from cookiecutter.exceptions import FailedHookException
from cookiecutter.generate import generate_files
from dict_tools.data import NamespaceDict


def __init__(hub):
    hub.pop.sub.add(dyne_name="tool")
    hub.pop.sub.load_subdirs(hub.tool, recurse=True)
    hub.pop.sub.load_subdirs(hub.pop_create)


COMPLETION_TEXT = """
Congratulations! You have set up a new project!
This project can be executed by calling the run.py script:
    python3 run.py
This project has been set up with pre-commit hooks for code checks and black.
First set up your source control environment with "git init" or "hg init".
Then enable these checks in your git checkout:
    pip install pre-commit
    pre-commit install
To run pre-commit manually, execute:
    pre-commit run --all-files
Please fork the pop-awesome and open a PR listing your new project \u263A
    https://gitlab.com/saltstack/pop/pop-awesome
"""


def cli(hub):
    hub.pop.config.load(["pop_create"], cli="pop_create")
    if hub.SUBPARSER:
        subparsers = [str(hub.SUBPARSER).replace("-", "_")]
    else:
        # No subparser specified, do all the core creators
        # cicd is last because it will run pre-commit on everything
        subparsers = ["seed", "docs", "tests", "cicd"]

    # Decide which copy function to use in the end
    if hub.OPT.pop_create.overwrite_existing:
        copy_function = hub.tool.path.copy2
    else:
        copy_function = hub.tool.path.copy

    with tempfile.TemporaryDirectory(prefix="pop-create-") as tempdir:
        ctx = hub.pop_create.init.context()
        hub.pop_create.init.run(
            directory=tempdir,
            subparsers=subparsers,
            root_dir=tempdir,
            **ctx,
        )

        hub.tool.path.IGNORE = [
            str(pathlib.Path(ctx.target_directory) / x)
            for x in hub.OPT.pop_create.ignore
        ]

        # Copy from temporary directory to target directory
        hub.tool.path.copytree(
            src=pathlib.Path(tempdir),
            dst=pathlib.Path(str(ctx.target_directory)),
            copy_function=copy_function,
        )

    # All done!
    print(COMPLETION_TEXT)


def context(hub):
    directory = pathlib.Path(hub.OPT.pop_create.directory)
    project_name = hub.OPT.pop_create.project_name or directory.name
    clean_name = project_name.replace("-", "_").replace(" ", "_")
    short_dyne_list = sorted(hub.OPT.pop_create.get("dyne", ()))
    if hub.OPT.pop_create.vertical:
        dyne_list = short_dyne_list
    else:
        dyne_list = [clean_name] + short_dyne_list

    if hub.OPT.pop_create.author:
        author = hub.OPT.pop_create.author
    else:
        author = subprocess.getoutput("git config --global user.name")

    if hub.OPT.pop_create.author_email:
        author_email = hub.OPT.pop_create.author_email
    else:
        author_email = subprocess.getoutput("git config --global user.email")

    ctx = NamespaceDict(
        author=author,
        author_email=author_email,
        clean_name=clean_name,
        dyne_list=dyne_list,
        project_name=project_name,
        short_dyne_list=short_dyne_list,
        this_year=str(datetime.datetime.today().year),
        target_directory=str(directory),
    )

    for key, value in hub.OPT.pop_create.items():
        # Don't overwrite any values we already sanitized
        if key in ("project_name", "directory", "author", "author_email"):
            continue
        ctx[key] = value
    return ctx


def run(
    hub,
    directory: pathlib.Path,
    subparsers: List[str],
    **ctx,
):
    for subparser in subparsers:
        try:
            # Let each subparser create it's own context
            context = NamespaceDict(ctx.copy())
            if "init" in hub.pop_create[subparser]:
                context = hub.pop_create[subparser].init.context(context, directory)

            # Get the input directory
            repo_dir = hub.pop_create[subparser]._dirs[0]

            # Copy the templates from the source to the a temporary directory using the generated context
            try:
                generate_files(
                    repo_dir=repo_dir,
                    context={"cookiecutter": context},
                    output_dir=directory,
                    # We should be dealing with a temporary directory
                    overwrite_if_exists=True,
                    skip_if_file_exists=False,
                )
            except FailedHookException:
                hub.log.error(f"Failed to run hook for subparser: {subparser}")
                raise
        except IndexError:
            hub.log.error(f"No template under sub {subparser}")



from unittest import mock
import pytest

@pytest.fixture(scope="session", name="hub")
def integration_hub(hub):
    for dyne in {{ cookiecutter.short_dyne_list if cookiecutter.vertical else cookiecutter.dyne_list }}:
        hub.pop.sub.add(dyne_name=dyne)

    with mock.patch("sys.argv", ["{{cookiecutter.get('entrypoint', cookiecutter.clean_name)}}"]):
        hub.pop.config.load(
            {{ cookiecutter.short_dyne_list if cookiecutter.vertical else cookiecutter.dyne_list }},
            cli="{{ cookiecutter.clean_name }}"
        )

    yield hub


async def _setup_hub(hub):
    # Set up the hub before each function here
    ...

async def _teardown_hub(hub):
    # Clean up the hub after each function here
    ...

@pytest.fixture(scope="function", autouse=True)
async def function_hub_wrapper(hub):
    await _setup_hub(hub)
    yield
    await _teardown_hub(hub)
